@@ -10,10 +10,12 @@ from .enums import (
     ENUM_D_TAG, ENUM_E_VERSION, ENUM_P_TYPE_BASE, ENUM_SH_TYPE_BASE,
     ENUM_RELOC_TYPE_i386, ENUM_RELOC_TYPE_x64,
     ENUM_RELOC_TYPE_ARM, ENUM_RELOC_TYPE_AARCH64, ENUM_RELOC_TYPE_PPC64,
-    ENUM_RELOC_TYPE_MIPS, ENUM_ATTR_TAG_ARM, ENUM_DT_FLAGS, ENUM_DT_FLAGS_1)
+    ENUM_RELOC_TYPE_MIPS, ENUM_ATTR_TAG_ARM, ENUM_ATTR_TAG_RISCV,
+    ENUM_RELOC_TYPE_S390X, ENUM_RELOC_TYPE_LOONGARCH, ENUM_DT_FLAGS,
+    ENUM_DT_FLAGS_1)
 from .constants import (
     P_FLAGS, RH_FLAGS, SH_FLAGS, SUNW_SYMINFO_FLAGS, VER_FLAGS)
-from ..common.py3compat import bytes2hex, iteritems
+from ..common.utils import bytes2hex
 
 
 def describe_ei_class(x):
@@ -149,8 +151,12 @@ def describe_reloc_type(x, elffile):
         return _DESCR_RELOC_TYPE_AARCH64.get(x, _unknown)
     elif arch == '64-bit PowerPC':
         return _DESCR_RELOC_TYPE_PPC64.get(x, _unknown)
+    elif arch == 'IBM S/390':
+        return _DESCR_RELOC_TYPE_S390X.get(x, _unknown)
     elif arch == 'MIPS':
         return _DESCR_RELOC_TYPE_MIPS.get(x, _unknown)
+    elif arch == 'LoongArch':
+        return _DESCR_RELOC_TYPE_LOONGARCH.get(x, _unknown)
     else:
         return 'unrecognized: %-7x' % (x & 0xFFFFFFFF)
 
@@ -194,7 +200,7 @@ def describe_ver_flags(x):
         VER_FLAGS.VER_FLG_INFO) if x & flag)
 
 
-def describe_note(x):
+def describe_note(x, machine):
     n_desc = x['n_desc']
     desc = ''
     if x['n_type'] == 'NT_GNU_ABI_TAG':
@@ -209,7 +215,7 @@ def describe_note(x):
     elif x['n_type'] == 'NT_GNU_GOLD_VERSION':
         desc = '\n    Version: %s' % (n_desc)
     elif x['n_type'] == 'NT_GNU_PROPERTY_TYPE_0':
-        desc = '\n      Properties: ' + describe_note_gnu_properties(x['n_desc'])
+        desc = '\n      Properties: ' + describe_note_gnu_properties(x['n_desc'], machine)
     else:
         desc = '\n      description data: {}'.format(bytes2hex(n_desc))
 
@@ -251,29 +257,26 @@ def describe_attr_tag_arm(tag, val, extra):
     else:
         return _DESCR_ATTR_TAG_ARM[tag] + d_entry[val]
 
+def describe_attr_tag_riscv(tag, val, extra):
+    idx = ENUM_ATTR_TAG_RISCV[tag] - 1
+    d_entry = _DESCR_ATTR_VAL_RISCV[idx]
 
-def describe_note_gnu_property_x86_feature_1(value):
+    if d_entry is None:
+        s = _DESCR_ATTR_TAG_RISCV[tag]
+        s += '"%s"' % val if val else ''
+        return s
+
+    else:
+        return _DESCR_ATTR_TAG_RISCV[tag] + d_entry[val]
+
+def describe_note_gnu_property_bitmap_and(values, prefix, value):
     descs = []
-    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS:
+    for mask, desc in values:
         if value & mask:
             descs.append(desc)
-    return 'x86 feature: ' + ', '.join(descs)
+    return '%s: %s' % (prefix, ', '.join(descs))
 
-def describe_note_gnu_property_x86_feature_2_used(value):
-    descs = []
-    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS:
-        if value & mask:
-            descs.append(desc)
-    return 'x86 feature used: ' + ', '.join(descs)
-
-def describe_note_gnu_property_x86_isa_1(value, verb):
-    descs = []
-    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS:
-        if value & mask:
-            descs.append(desc)
-    return 'x86 ISA %s: %s' % (verb, ', '.join(descs))
-
-def describe_note_gnu_properties(properties):
+def describe_note_gnu_properties(properties, machine):
     descriptions = []
     for prop in properties:
         t, d, sz = prop.pr_type, prop.pr_data, prop.pr_datasz
@@ -291,22 +294,27 @@ def describe_note_gnu_properties(properties):
             if sz != 4:
                 prop_desc = ' <corrupt length: 0x%x>' % sz
             else:
-                prop_desc = describe_note_gnu_property_x86_feature_1(d)
+                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS, 'x86 feature', d)
         elif t == 'GNU_PROPERTY_X86_FEATURE_2_USED':
             if sz != 4:
                 prop_desc = ' <corrupt length: 0x%x>' % sz
             else:
-                prop_desc = describe_note_gnu_property_x86_feature_2_used(d)                
+                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS, 'x86 feature used', d)                
         elif t == 'GNU_PROPERTY_X86_ISA_1_NEEDED':
             if sz != 4:
                 prop_desc = ' <corrupt length: 0x%x>' % sz
             else:
-                prop_desc = describe_note_gnu_property_x86_isa_1(d, "needed")
+                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS, 'x86 ISA needed', d)
         elif t == 'GNU_PROPERTY_X86_ISA_1_USED':
             if sz != 4:
                 prop_desc = ' <corrupt length: 0x%x>' % sz
             else:
-                prop_desc = describe_note_gnu_property_x86_isa_1(d, "used")
+                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS, 'x86 ISA used', d)
+        elif t == 'GNU_PROPERTY_AARCH64_FEATURE_1_AND' and machine == 'EM_AARCH64':
+            if sz != 4:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_AARCH64_FEATURE_1_AND, 'aarch64 feature', d)
         elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOPROC <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIPROC:
             prop_desc = '<processor-specific type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
         elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOUSER <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIUSER:
@@ -380,6 +388,7 @@ _DESCR_E_MACHINE = dict(
     EM_860='Intel 80860',
     EM_MIPS='MIPS R3000',
     EM_S370='IBM System/370',
+    EM_S390='IBM S/390',
     EM_MIPS_RS4_BE='MIPS 4000 big-endian',
     EM_IA_64='Intel IA-64',
     EM_X86_64='Advanced Micro Devices X86-64',
@@ -389,6 +398,8 @@ _DESCR_E_MACHINE = dict(
     EM_BLACKFIN='Analog Devices Blackfin',
     EM_PPC='PowerPC',
     EM_PPC64='PowerPC64',
+    EM_RISCV='RISC-V',
+    EM_LOONGARCH='LoongArch',
     RESERVED='RESERVED',
 )
 
@@ -410,7 +421,8 @@ _DESCR_P_TYPE = dict(
     PT_AARCH64_ARCHEXT='AARCH64_ARCHEXT',
     PT_AARCH64_UNWIND='AARCH64_UNWIND',
     PT_TLS='TLS',
-    PT_MIPS_ABIFLAGS='ABIFLAGS'
+    PT_MIPS_ABIFLAGS='ABIFLAGS',
+    PT_RISCV_ATTRIBUTES='RISCV_ATTRIBUT',
 )
 
 
@@ -450,6 +462,7 @@ _DESCR_SH_TYPE = dict(
     SHT_ARM_PREEMPTMAP='ARM_PREEMPTMAP',
     SHT_ARM_ATTRIBUTES='ARM_ATTRIBUTES',
     SHT_ARM_DEBUGOVERLAY='ARM_DEBUGOVERLAY',
+    SHT_RISCV_ATTRIBUTES='RISCV_ATTRIBUTES',
     SHT_MIPS_LIBLIST='MIPS_LIBLIST',
     SHT_MIPS_DEBUG='MIPS_DEBUG',
     SHT_MIPS_REGINFO='MIPS_REGINFO',
@@ -649,6 +662,11 @@ _DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS = (
     # TODO; there is a long list
 )
 
+# Same for GNU_PROPERTY_AARCH64_FEATURE_1_AND
+_DESCR_NOTE_GNU_PROPERTY_AARCH64_FEATURE_1_AND = (
+    (1, 'bti'),
+    (2, 'pac'),
+)
 
 def _reverse_dict(d, low_priority=()):
     """
@@ -660,7 +678,7 @@ def _reverse_dict(d, low_priority=()):
     not override any other entries of the same value.
     """
     out = {}
-    for k, v in iteritems(d):
+    for k, v in d.items():
         if v in out and k in low_priority:
             continue
         out[v] = k
@@ -671,7 +689,9 @@ _DESCR_RELOC_TYPE_x64 = _reverse_dict(ENUM_RELOC_TYPE_x64)
 _DESCR_RELOC_TYPE_ARM = _reverse_dict(ENUM_RELOC_TYPE_ARM)
 _DESCR_RELOC_TYPE_AARCH64 = _reverse_dict(ENUM_RELOC_TYPE_AARCH64)
 _DESCR_RELOC_TYPE_PPC64 = _reverse_dict(ENUM_RELOC_TYPE_PPC64)
+_DESCR_RELOC_TYPE_S390X = _reverse_dict(ENUM_RELOC_TYPE_S390X)
 _DESCR_RELOC_TYPE_MIPS = _reverse_dict(ENUM_RELOC_TYPE_MIPS)
+_DESCR_RELOC_TYPE_LOONGARCH = _reverse_dict(ENUM_RELOC_TYPE_LOONGARCH)
 
 _low_priority_D_TAG = (
     # these are 'meta-tags' marking semantics of numeric ranges of the enum
@@ -731,7 +751,6 @@ _DESCR_ATTR_TAG_ARM = dict(
     TAG_MPEXTENSION_USE_OLD='Tag_MPextension_use_old: ',
 )
 
-
 _DESCR_ATTR_VAL_ARM = [
     None, #1
     None, #2
@@ -984,3 +1003,27 @@ _DESCR_ATTR_VAL_ARM = [
         1: 'Allowed',
     },
 ]
+
+_DESCR_ATTR_TAG_RISCV = dict(
+    TAG_FILE='File Attributes',
+    TAG_SECTION='Section Attributes:',
+    TAG_SYMBOL='Symbol Attributes:',
+    TAG_STACK_ALIGN='Tag_RISCV_stack_align: ',
+    TAG_ARCH='Tag_RISCV_arch: ',
+    TAG_UNALIGNED='Tag_RISCV_unaligned_access: ',
+)
+
+_DESCR_ATTR_VAL_RISCV = [
+    None, #1
+    None, #2
+    None, #3
+    { #4 TAG_RISCV_stack_align
+        4: '4-bytes',
+        16: '16-bytes',
+    },
+    None, #5 TAG_RISCV_arch
+    { #6 TAG_RISCV_unaligned_access
+        0: 'Not Allowed',
+        1: 'Allowed',
+    },
+]
